/*
 * upd
 *
 * (C) 1989 Saeko Hirabauashi & Kouichi Hirabayashi
 *
 * Absolutely no warranty. Use this software with your own risk.
 *
 * Permission to use, copy, modify and distribute this software for any
 * purpose and without fee is hereby granted, provided that the above
 * copyright and disclaimer notice.
 */


#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#ifdef DOS
#include <process.h>
#endif
#include "upd.h"
#include "regexp.h"

extern char **FS, **RS, **OFS, **ORS, **FILENAME;
extern char **JFS, **JRS;
extern double *NF, *NR;
extern double *JNF, *JNR;
extern double *ARGC;
extern SYMBOL *argtab[];
extern CELL *getvar();

char *strsave(), *strcpy(), *getsval(), *jStrchr(), *strchr();
double getfval(), atof();
CELL *mkcell(), *mktmp(), *execute(), *patexec();
FILE *efopen();

static get1rec(), r_mkfld(), close1();
static char *updtmp();

extern CELL truecell, falsecell;

extern int pateval;

int infileno = 1;
FILE *ifp;
FILE *infile[2];
int mj;		/* 0: master, 1: jounal */
char record[2][RECSIZ];
CELL *field[2][MAXFLD];
int keyfld[2] = { 1, 1};
char cmptyp = 'l';	/* l: lexical, n: numerical */

char *fs_str;
regexp *fs_pat;

CELL *
Getline(p) NODE *p;
{
  CELL *u;
  char *fnam, *s, str[RECSIZ];
  int i;
  FILE *fp, *getfp();

  if ((int) p->n_arg[0])	/* read into var */
	s = str;
  else
	s = NULL;
  if ((int) p->n_arg[1]) {	/* file name */
	u = execute(p->n_arg[1]);
	fnam = getsval(u);
	fp = getfp(fnam, (int) p->n_arg[2]);
	c_free(u);
	i = get1rec(s, fp);
  }
  else
	i = get1rec(s, stdin);
  if (s == str) {
	u = execute(p->n_arg[0]);
	setsval(u, str);
  }
  return mktmp(NUM, NULL, (double) i);
}

static
get1rec(buf, fp) char *buf; FILE *fp;
{
  register int c;
  register char rs, *s;
  int mflg;

  if (buf == NULL)
	buf = record[0];
  if ((rs = **RS) == '\0') {	/* multi line record */
	mflg = 1;
	rs = '\n';
  }
  else
	mflg = 0;

  if (feof(fp) || (c = getc(fp)) == EOF)
	return 0;
  for (s = buf; ; ) {
	for ( ; c != rs && c != EOF; c = getc(fp)) {
		if (isKanji(c)) {
			*s++ = c; c = getc(fp);
		}
		*s++ = c;
	}
	if (mflg) {
		if ((c = getc(fp)) == '\n' || c == EOF)
			break;
		*s++ = '\n';
	}
	else
		break;
  }
  *s = '\0';
  if (buf == record[0]) {
	mkfld(record[0], *FS, field);
	(*NR)++;
  }
  return s > buf || c != EOF ? 1 : 0;
}

Getrec(mj)
{
  FILE *fp;
  register int c;
  register char *fs, rs, *s;
  char *buf;
  int mflg;

  buf = record[mj];
  fs = (mj) ? *JFS : *FS;
  rs = (mj) ? **JRS : **RS;
  if (rs == '\0') {	/* multi line record */
	mflg = 1;
	rs = '\n';
  }
  else
	mflg = 0;
  fp = infile[mj];

  if (feof(fp) || (c = getc(fp)) == EOF)
	return 0;
  for (s = buf; ; ) {
	for ( ; c != rs && c != EOF; c = getc(fp)) {
		if (isKanji(c)) {
			*s++ = c; c = getc(fp);
		}
		*s++ = c;
	}
	if (mflg) {
		if ((c = getc(fp)) == '\n' || c == EOF)
			break;
		*s++ = '\n';
	}
	else
		break;
  }
  *s = '\0';
  mkfld(record[mj], fs, field[mj]);
  if (mj)
	(*JNR)++;
  else
	(*NR)++;
  return s > buf || c != EOF ? 1 : 0;
}

mkfld(rec, sep, fld) char *rec, *sep; CELL *fld[];
{
  char *s, *t;
  char str[RECSIZ];
  int i, j, n;
  int skip = 0;
  double *nf;

  if (strlen(sep) > 1)
	return r_mkfld(rec, sep, fld);

  if (*sep == ' ' || *sep == '\0') {
	sep = " \t\n"; skip++;
  }
  nf = (rec == record[0]) ? NF : JNF;
  for (i = 1, n = (int)*nf; i <= n; i++) {
	sfree(fld[i]->c_sval);
	sfree(fld[i]);
	fld[i] = NULL;
  }
  for (i = 0, s = rec; ; ) {
	t = str;
	if (skip) {
		while (*s && strchr(" \t\n", *s))
			s++;
		if (*s == '\0')
			break;
	}
	while (*s && !jStrchr(sep, *s)) {
		if (isKanji(*s))
			*t++ = *s++;
		*t++ = *s++;
	}
	*t = '\0';
	fld[++i] =  mkcell(FLD|STR, str, 0.0);
	if (*s)
		s++;
	else
		break;
  }
  *nf = (double) i;
  return i;
}

static
r_mkfld(rec, sep, fld) char *rec, *sep; CELL *fld[];
{
  char *s, *t;
  char str[BUFSIZ];
  int i, n;
  regexp *mkpat();
  extern int r_start, r_length;

  if (strcmp(*FS, fs_str) != 0) {
	sfree(fs_str); sfree(fs_pat);
	fs_str = strsave(*FS);
	fs_pat = mkpat(fs_str);
  }
  for (i = 1, n = (int) *NF; i <= n; i++) {
	sfree(fld[i]->c_sval);
	sfree(fld[i]);
	fld[i] = NULL;
  }
  for (i = 0, s = rec, t = str; *s; ) {
	if (match(fs_pat, s)) {
		for (n = r_start; --n > 0; )
			*t++ = *s++;
	}
	else {
		while (*s)
			*t++ = *s++;
	}
	*t = '\0';
	t = str;
	fld[++i] = mkcell(FLD|STR, str, 0.0);
	if (*s)
		s += r_length;
  }
  *NF = (double) i;
  return i;
}

mkrec(u) CELL *u;
{
  register char *s, *t;
  register int i, j;
  double *nf;

  nf = (mj) ? JNF : NF;
  for (j = (int)*nf, i = 1; i <= j; i++)
	if (field[mj][i] == u)
		break;
  if (i > j) {
	for ( ; i < MAXFLD; i++)
		if (field[mj][i] == u)
			break;
	if (i == MAXFLD)
		error("too many field (%d)", i);
	*nf = (double)i;
  }
  for (t = record[mj], i = 1, j = (int)*nf; i <= j; i++) {
	if (i > 1)
		*t++ = **OFS;
	for (s = getsval(field[mj][i]); *s; )
		*t++ = *s++;
  }
  *t++ = '\0';
}

CELL *
Field(p) NODE *p;
{
  CELL *u;
  int i, j, m;

  u = execute(p->n_arg[0]);
  i = (int) getfval(u);
  c_free(u);
  m = ((int)p->n_arg[1] == '$') ? 0 : 1;
  j = (m) ? (int)*JNF : (int)*NF;
  if (i > j)
	for (++j; j <= i; j++) {
		if (field[m][j] == NULL)
			field[m][j] = mkcell(FLD|STR, "", 0.0);
	}
  u = field[m][i];
  if (u == (CELL*)NULL)
	u = &falsecell;
  return u;
}

CELL *
Print0()
{
#if 1
  int i, j;
  char *s, str[BUFSIZ];

  for (*str = '\0', i = 1, j = (int) *NF; i <= j; i++) {
	if (i > 1)
		strcat(str, *OFS);
	s = getsval(field[i]);
	strcat(str, s);
  }
  strcat(str, *ORS);
  fputs(str, stdout);
#else
  fprintf(stdout, "%s%s", record, *ORS);
#endif
  return &truecell;
}

char *
format(t, p) char *t; NODE *p;
{
  CELL *u, *v;
  char *r, *s, *s0, fmt[BUFSIZ];
  double x;
  int i;

  u = execute(p->n_arg[2]);
  s = s0 = getsval(u);
  for (i = 3; *s; s++) {
	if (isKanji(*s)) {
		*t++ = *s++; *t++ = *s; continue;
	}
	if (*s != '%') {
		*t++ = *s; continue;
	}
	else if (*(s + 1) == '%') {
		*t++ = *s++; continue;
	}
	for (r = fmt, *r++ = *s++; *r++ = *s; s++) {
		if (strchr("%cdefgosux", *s))
			break;
	}
	*r = '\0';
	if (p->n_arg[i] == NULL)
		error("not enough args in printf(%s)", s0);
	v = execute(p->n_arg[i++]);
	if (*s == 's')
		r = getsval(v);
	else
		x = getfval(v);
	switch (*s) {
	case 'c':
		sprintf(t, fmt, (int) x);
		break;
	case 'd':
		if (*(s - 1) != 'l') {
			*--r = 'l'; *++r = 'd'; *++r = '\0';
		}
		sprintf(t, fmt, (long) x);
		break;
	case 'e': case 'f': case 'g':
		sprintf(t, fmt, x);
		break;
	case 'o': case 'u': case 'x':
		if (*(s - 1) == 'l')
			sprintf(t, fmt, (long) x);
		else
			sprintf(t, fmt, (int) x);
		break;
	case 's':
		/*r = getsval(v);*/
		sprintf(t, fmt, r);
		break;
	default:
		strcpy(t, fmt);
		break;
	}
	c_free(v);
	t += strlen(t);
  }
  c_free(u);
  *t = '\0';
}

#define MAXFILE	10
struct {
  char *f_name;	/* file name */
  FILE *f_fp;
  int f_type;
} filetab[MAXFILE];

FILE *
getfp(file, type) char *file;
{
  register int i;
  register char *name, *mode;
  FILE *fp, *efopen(), *epopen();

  for (i = 0; i < MAXFILE; i++)
	if (filetab[i].f_name && strcmp(filetab[i].f_name, file) == 0)
		return filetab[i].f_fp;
  for (i = 0; i < MAXFILE; i++)
	if (!filetab[i].f_fp)
		break;
  if (i == MAXFILE)
	error("too many files to open");
  name = file;
  switch (type) {
  case R_OUT:	mode = "w"; break;
  case R_APD:	mode = "a"; break;
  case R_POUT:
	fp = epopen(file, "w");
	goto g1;
	break;
  case R_IN:	mode = "r"; break;
  case R_PIN:
	fp = epopen(file,"r");
	goto g1;
	break;
  }
  fp = efopen(name, mode);
g1:
  filetab[i].f_name = strsave(file);
  filetab[i].f_type = type;
  filetab[i].f_fp = fp;
  return fp;
}

closeall()
{
  register int i;

  for (i = 0; i < MAXFILE; i++)
	close1(i);
}

CELL *
Close(s) char *s;
{
  register int i;

  for (i = 0; i < MAXFILE; i++)
	if (strcmp(s, filetab[i].f_name) == 0) {
		close1(i);
		break;
	}
  i = (i == MAXFILE) ? 0 : 1;
  return mktmp(NUM, NULL, (double) i);
}

static
close1(i)
{
  int fd, result, savefd;
  char *updtmp();

  if (filetab[i].f_fp == NULL)
	return;
  switch (filetab[i].f_type) {
  case R_PIN:
	pclose(filetab[i].f_fp);
	break;
  case R_IN: case R_OUT: case R_APD:
	fclose(filetab[i].f_fp);
	break;
  case R_POUT:
	pclose(filetab[i].f_fp);
	break;
  }
  sfree(filetab[i].f_name);
  filetab[i].f_type = 0;
  filetab[i].f_name = NULL;
  filetab[i].f_fp = NULL;
}

FILE *
epopen(file, mod) char *file, *mod;
{
  FILE *fp, *popen();

  if ((fp = popen(file, mod)) == NULL)
	error("can't poen %s", file);
  return fp;
}

static char *
updtmp(i)
{
  static char str[16];

  sprintf(str, "upd000%02d.tmp", i);
  return str;
}

Index(s, t) char *s, *t;
{
  register char *u, *v;
  register int i;

  for (i = 1; *s; s++, i++) {
	for (u = s, v = t; *v; u++, v++) {
		if (isKanji(*v)) {
			if (*u != *v)
				break;
			u++; v++;
		}
		if (*u != *v)
			break;
	}
	if (*v == '\0')
		return i;
	if (isKanji(*s))
		s++;
  }
  return 0;
}

keycmp()
{
  char *s, *t;
  double cond, x, y, getfval();
  int r;

  if ((int)*NF < keyfld[0])
	error("no key field in master file", (char *)0);;
  if ((int)*JNF < keyfld[1])
	error("no key field in transaction file", (char *)0);;
  if (cmptyp != 'n') {
	s = getsval(field[0][keyfld[0]]);
	t = getsval(field[1][keyfld[1]]);
	r = strcmp(s, t);
  }
  else {
	x = getfval(field[0][keyfld[0]]);
	y = getfval(field[1][keyfld[1]]);
	cond = x - y;
	r = 0;
	if (cond < 0.0)
		r = -1;
	else if (cond > 0.0)
		r = 1;
  }
  return r;
}
